#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock2.h>
#include <stdio.h>
#include <process.h>

#define NOMBRE_SERVICIO "runserv"

#define NADA		0
#define CREAR		1
#define BORRAR	2

#define MAX_APL		 1
#define DELAY			 5

#define PUERTO	8000

SERVICE_STATUS_HANDLE sshStatusHandle;
SERVICE_STATUS ssStatus;
HANDLE  hServerStopEvent = NULL;
HANDLE	hFile;

BOOL bResult;
HWND hWin[MAX_APL];
char Path[MAX_PATH], szBufferError[MAX_PATH];
int  iHwnd;

DWORD dwError, dwPIDObjetivo;

HANDLE  hApp;

void BucleMensajes()
{
	MSG Message;

	while(PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);		
	}
	return;
}

HANDLE LaunchApp (LPCTSTR lpCmdLine, LPCTSTR szPath, LPTSTR szDominio, LPTSTR szUser, LPTSTR szPassword, BOOL bVisible, BOOL bSincrono, int iTimeout)
{
  STARTUPINFO         si;
  BOOL                bOk;
  PROCESS_INFORMATION pi;
  int i;
	DWORD exitCode;
	HANDLE	hUser;

	// Sets the current directory if path input is not empty
	if(strlen(szPath) != 0)
	{
		SetCurrentDirectory(szPath);
	}

	if(strlen(szDominio) == 0)
	{
		szDominio = NULL;
	}

	// Logon with User and Password specified
	bOk = LogonUser(szUser,
									szDominio,
									szPassword,
									LOGON32_LOGON_INTERACTIVE,
									LOGON32_PROVIDER_DEFAULT,
									&hUser);

	if(bOk == FALSE)
	{
		dwError = GetLastError();
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
									NULL,
									dwError,
									MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
									(LPTSTR) &szBufferError,
									MAX_PATH,
									NULL);
		
		return (HANDLE)NULL;
	}

  ZeroMemory (&pi, sizeof(pi));
  ZeroMemory (&si, sizeof(si));
  
	si.cb = sizeof(STARTUPINFO);
	si.lpReserved = NULL;
	si.lpTitle = NULL;
	si.lpDesktop = "WinSta0\\Default";
	si.dwX = si.dwY = si.dwXSize = si.dwYSize = 0L;
	si.dwFlags = STARTF_USESHOWWINDOW;

	if(bVisible == TRUE)
	{
		si.wShowWindow = SW_NORMAL;
	}
	else
	{
		si.wShowWindow = SW_HIDE;
	}

	si.lpReserved2 = NULL;
	si.cbReserved2 = 0;
	
	// Create process with user connected
	bOk = CreateProcessAsUser(hUser, NULL, (LPSTR)lpCmdLine, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi);
	if(!bOk)
	{
		dwError = GetLastError();
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
									NULL,
									dwError,
									MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
									(LPTSTR) &szBufferError,
									MAX_PATH,
									NULL);

		return (HANDLE)NULL;
	}
  else
	{
		// Obtains PID of process started
		dwPIDObjetivo = pi.dwProcessId;

		// Checks if mode is synchronous or not for waiting the end proccess
		if(bSincrono == TRUE)
		{
			if(iTimeout == -1)
			{
				// Synchronous with timeout infinite
				while((GetExitCodeProcess((HANDLE)pi.hProcess, &exitCode)) && 
							(exitCode == STILL_ACTIVE))
				{
					// "Cleans" the message queue
					BucleMensajes();
					Sleep(1000);
				}  // while
			}
			else
			{
				// Synchronous with timeout
				i = 0;

				while((GetExitCodeProcess((HANDLE)pi.hProcess, &exitCode)) && 
							(exitCode == STILL_ACTIVE) &&
							(i < iTimeout))
				{
					// "Cleans" the message queue
					BucleMensajes();

					Sleep(1000);
					i++;
				} // while
			} // if(Timeout == -1)  
		} // if(bSincrono == TRUE)

		return pi.hProcess;
	} // if(!bOk)
} // Fin de  LaunchApp



BOOL ReportStatusToSCMgr(DWORD dwCurrentState, DWORD dwWin32ExitCode, DWORD dwWaitHint)
{
	static DWORD dwCheckPoint = 1;
	BOOL fResult = TRUE;

	if(dwCurrentState == SERVICE_START_PENDING)
	{
		ssStatus.dwControlsAccepted = 0;
	}
	else
	{
		ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	}

	ssStatus.dwCurrentState = dwCurrentState;
	ssStatus.dwWin32ExitCode = dwWin32ExitCode;
	ssStatus.dwWaitHint = dwWaitHint;

	if((dwCurrentState == SERVICE_RUNNING) || (dwCurrentState == SERVICE_STOPPED))
	{
		ssStatus.dwCheckPoint = 0;
	}
	else
	{
		ssStatus.dwCheckPoint = dwCheckPoint++;
	}

	// Updates the service status
	return SetServiceStatus(sshStatusHandle, &ssStatus);
}


VOID ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv)
{
	// Updates the service status
	bResult = ReportStatusToSCMgr(SERVICE_START_PENDING, NO_ERROR, (DELAY+1)*1000);

	// Updates the service status
	bResult = ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 0);

	return;
}

VOID ServiceStop()
{
	// Updates the service status to pending to stop
	bResult = ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, (DELAY+1)*1000);

	// // Updates the service status to stop
	bResult = ReportStatusToSCMgr(SERVICE_STOPPED, NO_ERROR, 0);

	return;
}


VOID WINAPI service_ctrl(DWORD dwCtrlCode)
{
	// Handle the requested control code.
	switch(dwCtrlCode)
	{
		case SERVICE_CONTROL_STOP:
			
			// Stops the service
			if(hServerStopEvent != NULL)
			{
				bResult = SetEvent(hServerStopEvent);
			} // if(hServerStopEvent != NULL)

			break;
		
		default:
			
			break;
	}
}



void WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)
{
	HANDLE					hEvents[1] = {NULL};
  DWORD						dwWait;
	char						szComando[MAX_PATH], szPath[MAX_PATH], szModo[MAX_PATH], szDominio[MAX_PATH];
	char						szServidor[MAX_PATH], szUsuario[MAX_PATH], szPassword[MAX_PATH];
	char						szBuffer[4*MAX_PATH];
	BOOL						bVisible;
	int							i, j, ioctl_opt, iError;
	struct					sockaddr_in dirRecepcion;
	WSADATA					wsaData;
	SOCKET					idSocket, idSocketAcepta;
  SOCKADDR_IN			acc_sin;
  int							acc_sin_len;
	char						*interface = NULL;

	ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	ssStatus.dwServiceSpecificExitCode = 0;

	// Register the function with the service
	sshStatusHandle = RegisterServiceCtrlHandler(TEXT(NOMBRE_SERVICIO), service_ctrl);
  if(sshStatusHandle != 0)
  {
		
		// Create the stop event service
		hServerStopEvent = CreateEvent(NULL,
																	 TRUE,
																	 FALSE,
																	 NULL);

		if(hServerStopEvent != NULL)
		{
			// Event created correctly
			hEvents[0] = hServerStopEvent;

			// Starts the service
			ServiceStart(dwArgc, lpszArgv);

			// Initialize WS2_32.DLL librarie
			if(WSAStartup(0x202, &wsaData) == SOCKET_ERROR)
			{
				iError = WSAGetLastError();
				sprintf(szBufferError, "Error %d in WSAStartup()\n", iError);
				WSACleanup();

				// Stop the service
				ServiceStop();
				return;
			}

			// Creates socket
			idSocket = socket(AF_INET, SOCK_STREAM, 0);
			if(idSocket == INVALID_SOCKET)
			{
				iError = WSAGetLastError();
				sprintf(szBufferError, "Error %d in socket()\n", iError);
				WSACleanup();

				// Stop the service
				ServiceStop();
				return;
			}

			// Obtains the local machine name
			if(gethostname(szServidor, MAX_PATH) == SOCKET_ERROR)
			{
				iError = WSAGetLastError();
				sprintf(szBufferError, "Error %d in gethostname()\n", iError);
				closesocket(idSocket);
				WSACleanup();

				// Stop the service
				ServiceStop();
				return;
			};

			dirRecepcion.sin_addr.s_addr = INADDR_ANY;
			dirRecepcion.sin_family = AF_INET;
			dirRecepcion.sin_port = htons(PUERTO);

			// Bind socket with IP address and port
			if(bind(idSocket, (struct sockaddr FAR *)&dirRecepcion, sizeof(dirRecepcion)) == SOCKET_ERROR)
			{
				iError = WSAGetLastError();
				sprintf(szBufferError, "Error %d in bind()\n", iError);
				closesocket(idSocket);
				WSACleanup();

				// Stop the service
				ServiceStop();
				return;
			}

			// Sets non-blocking mode
			ioctl_opt = 1;
			if(ioctlsocket(idSocket, FIONBIO, &ioctl_opt) == SOCKET_ERROR)
			{ 
				iError = WSAGetLastError();
				sprintf(szBufferError, "Error %d in ioctlsocket()\n", iError);
				closesocket(idSocket);
				WSACleanup();

				// Stop the service
				ServiceStop();
				return;
			}

			// Waits the stop event
			while(1)
			{
				// Waits the stop event or one second
				// Checks if there is information
				memset(szComando, 0, MAX_PATH);
				memset(szPath, 0, MAX_PATH);
				memset(szModo, 0, MAX_PATH);
				memset(szUsuario, 0, MAX_PATH);
				memset(szPassword, 0, MAX_PATH);
				memset(szDominio, 0, MAX_PATH);
				memset(szBuffer, 0, 4*MAX_PATH);

				// Waits a connection
				if(listen(idSocket, SOMAXCONN) == SOCKET_ERROR)
				{ 
					iError = WSAGetLastError();
					sprintf(szBufferError, "Error %d in listen()\n", iError);
					closesocket(idSocket);
					WSACleanup();

					// Stops the service
					ServiceStop();
					return;
				}

				// Accept a connection
				acc_sin_len = sizeof(acc_sin);
				idSocketAcepta = accept(idSocket, (struct sockaddr FAR *) &acc_sin, (int FAR *)&acc_sin_len);
				if(idSocketAcepta == INVALID_SOCKET)
				{ 
					// Error in accept. There is not a connection.
					iError = WSAGetLastError();
					if(iError == WSAEWOULDBLOCK)
					{
						// There is not any pending connection. don't do anything
					}
					else
					{
						// Error is different of there is not any pending connection
						iError = WSAGetLastError();
						sprintf(szBufferError, "Error %d in accept()\n", iError);
						closesocket(idSocket);
						WSACleanup();

						// Stops the service
						ServiceStop();
						return;
					} // if(rc != WSAEWOULDBLOCK)
				}
				else
				{
					// There is connection. Proceed
					// Gets data received
					memset(szBuffer, 0, MAX_PATH);
					iError = recv(idSocketAcepta, szBuffer, MAX_PATH, 0);
					if(iError == SOCKET_ERROR)
					{
						// Error in recepcin
						iError = WSAGetLastError();
						if(iError == WSAEWOULDBLOCK)
						{
							// There is not anything to read. Dont do anything
						}
						else
						{
							// Another error
							sprintf(szBufferError, "Error %d in recv()\n", iError);
						}// if(WSAGetLastError() == WSAEWOULDBLOCK)
					}
					else
					{	
						// Message received correctly. Run command
						for(i=0; (szBuffer[i]!=0xa) && (i<(int)strlen(szBuffer)); i++)
						{
							szComando[i] = szBuffer[i];
						}

						for(j=i+1; (szBuffer[j]!=0xa) && (j<(int)strlen(szBuffer)); j++)
						{
							szPath[j-i-1] = szBuffer[j];
						}

						for(i=j+1; (szBuffer[i]!=0xa) && (i<(int)strlen(szBuffer)); i++)
						{
							szModo[i-j-1] = szBuffer[i];
						}

						for(j=i+1; (szBuffer[j]!=0xa) && (j<(int)strlen(szBuffer)); j++)
						{
							szUsuario[j-i-1] = szBuffer[j];
						}

						for(i=j+1; (szBuffer[i]!=0xa) && (i<(int)strlen(szBuffer)); i++)
						{
							szPassword[i-j-1] = szBuffer[i];
						}

						for(j=i+1; (szBuffer[j]!=0xa) && (j<(int)strlen(szBuffer)); j++)
						{
							szDominio[j-i-1] = szBuffer[j];
						}

						_strupr(szComando);
						_strupr(szPath);
						_strupr(szModo);
						_strupr(szDominio);
						_strupr(szUsuario);
						
						if(strlen(szComando) != 0)
						{
							// Run command
							if(strlen(szModo) == 0)
							{
								bVisible = FALSE;
							}
							else
							{
								if(stricmp(szModo, "OCULTO") == 0)
								{
									bVisible = FALSE;
								}
								else
								{
									bVisible = TRUE;
								}
							} // if(strlen(szModo) == 0)

							// Dontt run any command related with runserv
							if(strstr(szComando, "runserv") == NULL)
							{
								if(LaunchApp(szComando, szPath, szDominio,
														 szUsuario, szPassword, bVisible, TRUE, 60) != NULL)
								{
									// Answer client request
									// Builds message to send
									memset(szBuffer, 0, MAX_PATH);
									strcpy(szBuffer, "OK");
									szBuffer[strlen(szBuffer)] = 0xa;

									if(send(idSocketAcepta, szBuffer, strlen(szBuffer), 0) == SOCKET_ERROR)
									{
										iError = WSAGetLastError();
										sprintf(szBufferError, "Error %d in send()\n", iError);
									}
								}
								else
								{
									// Send error answer to client
									memset(szBuffer, 0, MAX_PATH);
									strcpy(szBuffer, "ERROR");
									szBuffer[strlen(szBuffer)] = 0xa;

									// Send error message
									strcat(szBuffer, szBufferError);
									szBuffer[strlen(szBuffer)] = 0xa;

									if(send(idSocketAcepta, szBuffer, strlen(szBuffer), 0) == SOCKET_ERROR)
									{
										iError = WSAGetLastError();
										sprintf(szBufferError, "Error %d en send()\n", iError);
									}
								} // if(LaunchApp...
							} // if(strstr(szComando, "runserv") == NULL)
						}
						else
						{
							// There is not any command to run
						} // if(strlen(szComando) != 0)
					} // if((recv(idSocket, szBuffer, MAX_PATH, 0)) != SOCKET_ERROR)
	
					// Closes socket
					closesocket(idSocketAcepta);
				} // if(idSocketAcepta == INVALID_SOCKET)

				// Wait one second before next loop
				Sleep(1000);
	
				dwWait = WaitForMultipleObjects(1, hEvents, FALSE, 1000);

				// If stop event ocurred, ends loop
				if(dwWait - WAIT_OBJECT_0 == 0) break;
			} // while(1)

			// Stops the service
			ServiceStop();
		} // if(hServerStopEvent != NULL)
	} // if(sshStatusHandle != 0)

	// Closes stop event handle
	if(hServerStopEvent != NULL)
	{
		CloseHandle(hServerStopEvent);
	}

 	// Frees WS2_32.DLL librarie
	closesocket(idSocket);
	WSACleanup();

	return;
}

void Ayuda(void)
{
	system("cls");
	printf("Syntax: runserv <Option> [IP address]\n");
	printf("        [-c]: Creates service runserv.\n");
	printf("        [-b]: Erases service runserv.\n");
	printf("Wait...\n");

	return;
}

/*******************************************************************/
// Main program

int main(int argc, char **argv)
{
	char	szPath[MAX_PATH];

	int iOpcion = NADA;
	int i;

  SC_HANDLE   schService;
  SC_HANDLE   schSCManager;

  SERVICE_TABLE_ENTRY dispatchTable[] =
  {
    {TEXT(NOMBRE_SERVICIO), (LPSERVICE_MAIN_FUNCTION)service_main},
    {NULL, NULL}
  };

	// Reads program parameters
	if(argc < 2)
	{
		// No parameters. Calls StartServiceCtrlDispatcher
		Ayuda();

		bResult = StartServiceCtrlDispatcher(dispatchTable);

		if(bResult == FALSE)
		{
			// Error in StartServiceCtrlDispatcher
			dwError = GetLastError();
			exit((int)dwError);
		}
		else
		{
			// StartServiceCtrlDispatcher correct
			exit(0);
		} // if(bResult == FALSE)
	} // if(argc < 2)

	// Reads parameters
	for(i=1; i < argc; i++)
	{
		if(stricmp(argv[i], "-c") == 0)
		{
			if(iOpcion == NADA)
			{
				iOpcion = CREAR;
			}
			else
			{
				iOpcion = NADA;
			} // if(iOpcion == NADA)
		}
		else if(stricmp(argv[i], "-b") == 0)
		{
			if(iOpcion == NADA)
			{
				iOpcion = BORRAR;
			}
			else
			{
				iOpcion = NADA;
			} // if(iOpcion == NADA)
		} // if(stricmp(argv[i]...
	} // for(i=1; i < argc; i++)

	switch(iOpcion)
	{
		case CREAR:
		{
			// Installs service
			schSCManager = OpenSCManager(NULL,										// machine (NULL == local)
																	 NULL,										// database (NULL == default)
																	 SC_MANAGER_ALL_ACCESS);	// access required
			if(schSCManager != NULL)
			{
				// OpenSCManager OK
				GetModuleFileName(NULL, szPath, MAX_PATH);

				schService = CreateService(schSCManager,               // SCManager database
																	 TEXT(NOMBRE_SERVICIO),	     // name of service
																	 TEXT(NOMBRE_SERVICIO),			 // name to display
																	 SERVICE_ALL_ACCESS,         // desired access
																	 SERVICE_WIN32_OWN_PROCESS |
																	 SERVICE_INTERACTIVE_PROCESS,// service type
																	 SERVICE_AUTO_START,         // start type
																	 SERVICE_ERROR_NORMAL,       // error control type
																	 szPath,                     // service's binary
																	 NULL,                       // no load ordering group
																	 NULL,                       // no tag identifier
																	 NULL,								       // dependencies
																	 NULL,					             // LocalSystem account
																	 NULL);                      // no password

				if(schService != NULL)
				{
					// CreateService OK
					// Starts service
					if(StartService(schService, 0, NULL) == TRUE)
					{
						// Service started
						CloseServiceHandle(schService);
						dwError = 0;
					}
					else
					{
						// Error in StartService
						dwError = GetLastError();
						printf("Error %d in StartService()\n", dwError);
					}
				}
				else
				{
					// Error in CreateService
					dwError = GetLastError();
					printf("Error %d in CreateService()\n", dwError);
				} // if(schService != NULL)
      
				CloseServiceHandle(schSCManager);
			}
			else
			{
				// Error in OpenSCManager
				dwError = GetLastError();
			} // if(schSCManager != NULL)
	
			break;
		} // case CREAR:
	
		case BORRAR:
		{
			// Erases service
	    schSCManager = OpenSCManager(NULL,NULL,SC_MANAGER_ALL_ACCESS);
			if(schSCManager != NULL)
			{
				schService = OpenService(schSCManager, TEXT(NOMBRE_SERVICIO), SERVICE_ALL_ACCESS);
				if(schService != NULL)
				{
					// Checks service status
					QueryServiceStatus(schService, &ssStatus);
					if(ssStatus.dwCurrentState == SERVICE_STOPPED)
					{
						// Service is stopped. Tries to erase it
						if(DeleteService(schService))
						{
							// Service deleted
							dwError = 0;
						}
						else
						{
							// Error in DeleteService
							dwError = GetLastError();
							printf("Error %d in DeleteService()\n", dwError);
						} // if(DeleteService(schService))
						
						CloseServiceHandle(schService);

						break;
					}
					else
					{
						// Service is not stopped. Tries to stop it
						if(ControlService(schService, SERVICE_CONTROL_STOP, &ssStatus) != 0)
						{
							while(QueryServiceStatus(schService, &ssStatus))
							{
								if(ssStatus.dwCurrentState == SERVICE_STOP_PENDING)
								{
									Sleep(1000);
								}
								else
								{
									break;
								} // if(ssStatus.dwCurrentState == SERVICE_STOP_PENDING)
							} // while(QueryServiceStatus(schService, &ssStatus))
							
							if(ssStatus.dwCurrentState == SERVICE_STOPPED)
							{
								// Service stopped
							}
							else
							{
								// Service is not stopped
								dwError = GetLastError();
								printf("Error %d stopping service runserv\n", dwError);
							} // if(ssStatus.dwCurrentState == SERVICE_STOPPED)
							
							// Erases service
							if(DeleteService(schService))
							{
								// Service deleted
								dwError = 0;
							}
							else
							{
								// Error in DeleteService
								dwError = GetLastError();
								printf("Error %d in DeleteService()\n", dwError);
							} // if(DeleteService(schService))
							
							CloseServiceHandle(schService);
						}
						else
						{
							// Error in ControlService
							dwError = GetLastError();
							printf("Error %d in ControlService()\n", dwError);
						} // if(ControlService(schService, SERVICE_CONTROL_STOP, &ssStatus))
					} // if(ssStatus.dwCurrentState == SERVICE_STOPPED)
				}
				else
				{
					// Error in OpenService
					dwError = GetLastError();
					printf("Error %d in OpenService()\n", dwError);
				} // if(schService != NULL)

				CloseServiceHandle(schSCManager);
			}
			else
			{
				// Error in OpenSCManager
				dwError = GetLastError();
			} // if(schService != NULL)

			break;
		} // case BORRAR:

		default:
		{
			exit(-1);
			break;
		}
	} // switch(iOpcion)

	exit((int)dwError);
}